# -*- coding: utf-8 -*-
"""define some const terms
Author  : NextPCG
"""

dson_inputs_tag = "Inputs"
dson_input_tag = "input"
dson_params_tag = "Params"
dson_outputs_tag = "Outputs"
dson_output_tag = "output"
dson_doc_tag = "Doc"
dson_meta_field_tag = "DsonMeta"
dson_meta_clsname_tag = "clsName"
dson_meta_if_in_server_tag = "ifInServer"
dson_meta_plugin_name = "pluginName"
